#
# Copyright (c) 1997 Regents of the University of California.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
# 	This product includes software developed by the MASH Research
# 	Group at the University of California Berkeley.
# 4. Neither the name of the University nor of the Research Group may be
#    used to endorse or promote products derived from this software without
#    specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Code contributed by Giao Nguyen, http://daedalus.cs.berkeley.edu/~gnguyen
#

# Copyright (c) 2002 Massachusetts Institute of Technology,
# All rights reserved.
# 
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose, without fee, and without
# written agreement is hereby granted, provided that the above
# copyright notice and the following two paragraphs appear in all
# copies of this software.
# 
# IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE
# LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
# CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND
# ITS DOCUMENTATION, EVEN IF THE MASSACHUSETTS INSTITUTE OF
# TECHNOLOGY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY SPECIFICALLY DISCLAIMS
# ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
# SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE
# MASSACHUSETTS INSTITUTE OF TECHNOLOGY HAS NO OBLIGATION TO
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
# 
# Code contributed by Ronny Krashinsky, ronny@mit.edu, http://www.cag.lcs.mit.edu/~ronny

proc rvValue {rv {op ""}} {
	if [catch "$rv value" val] {
		set val $rv
	}
	if {$op == ""} {
		return $val
	} else {
		return  [expr ${op}($val)]
	}
}


Class Http

Http set srcType_ TCP/Reno
Http set snkType_ TCPSink
Http set phttp_ 0
Http set maxConn_ 4

Http set tr_base_ data/http
Http set max_page_ 100

# The following rv* parameters can take a value or a RandomVariable object
Http set rvClientTime_ 3.0
Http set rvServerTime_ 0
Http set rvReqLen_ 256
Http set rvRepLen_ 4000
Http set rvImgLen_ 8000
Http set rvNumImg_ 4

Http instproc srcType {val} { $self set srcType_ $val }
Http instproc snkType {val} { $self set snkType_ $val }
Http instproc phttp {val} { $self set phttp_ $val }
Http instproc maxConn {val} { $self set maxConn_ $val }
Http instproc tr_base {val} { $self set tr_base_ $val }
Http instproc max_page {val} { $self set max_page_ $val }
Http instproc rvClientTime {val} { $self set rvClientTime_ $val }
Http instproc rvServerTime {val} { $self set rvServerTime_ $val }
Http instproc rvReqLen {val} { $self set rvReqLen_ $val }
Http instproc rvRepLen {val} { $self set rvRepLen_ $val }
Http instproc rvImgLen {val} { $self set rvImgLen_ $val }
Http instproc rvNumImg {val} { $self set rvNumImg_ $val }

Http proc setCDF {rvName value} {
	if [file exists $value] {
		set rv [new RandomVariable/Empirical]
		$rv loadCDF $value
		set value $rv
	}
	eval "$self set ${rvName}_ $value"
	return $value
}


Http instproc init {ns client server args} {
    eval $self init-vars $args

    $self instvar srcType_ snkType_ agents_ sinks_
    $self instvar phttp_ maxConn_ rvClientTime_ rvServerTime_ pagecount_ max_page_
    $self instvar rvReqLen_ rvRepLen_ rvNumImg_ rvImgLen_
    $self instvar ns_ numImg_ numGet_ numPut_ client_ server_ tStart_
    $self instvar tr_base_ rv_reqlen_tr_ rv_replen_tr_ rv_imglen_tr_ rv_imgnum_tr_ rv_ctime_tr_ rv_stime_tr_ 
    $self instvar pagetime_tr_
    

    set ns_ $ns
    set client_(node) $client
    set server_(node) $server
    if {$phttp_} {
	set $maxConn_ 1
    }

    set pagecount_ 0

    set rv_reqlen_tr_ [open "$tr_base_-reqlen.dat" w]
    set rv_replen_tr_ [open "$tr_base_-replen.dat" w]
#    set rv_imglen_tr_ [open "$tr_base_-imglen.dat" w]
    set rv_imgnum_tr_ [open "$tr_base_-imgnum.dat" w]
    set rv_ctime_tr_ [open "$tr_base_-ctime.dat" w]
    set rv_stime_tr_ [open "$tr_base_-stime.dat" w]
    set pagetime_tr_ [open "$tr_base_-pagetime.dat" w]
    
    
    for {set i 0} {$i < $maxConn_} {incr i} {
	set client_tcp [new Agent/TCP/FullTcp]
	set server_tcp [new Agent/TCP/FullTcp]
	lappend agents_(source) $client_tcp $server_tcp
	
	$client_tcp proc done {} "$self client-done $i"
	$server_tcp proc done {} "$self server-done $i"
	
	$server_tcp listen
	
	$ns_ attach-agent $client $client_tcp
	$ns_ attach-agent $server $server_tcp
	$ns_ connect $client_tcp $server_tcp
	$ns_ connect $server_tcp $client_tcp
	
	set client_app [new Application]
	$client_app proc recv {bytes} "$self client-recv $i \$bytes"
	$client_app attach-agent $client_tcp
	set client_($i) $client_app
	
	set server_app [new Application]
	$server_app proc recv {bytes} "$self server-recv $i \$bytes"
	$server_app attach-agent $server_tcp
	set server_($i) $server_app

	set sinks_($client_($i)) $server_tcp
	set sinks_($server_($i)) $client_tcp

	$client_app start
	$server_app start
    }
}

Http instproc finish {} {
    $self instvar rv_reqlen_tr_ rv_replen_tr_ rv_imglen_tr_ rv_imgnum_tr_ rv_ctime_tr_ rv_stime_tr_
    $self instvar pagetime_tr_
    
    close $rv_reqlen_tr_
    close $rv_replen_tr_
#    close $rv_imglen_tr_
    close $rv_imgnum_tr_
    close $rv_ctime_tr_
    close $rv_stime_tr_
    close $pagetime_tr_

}

Http instproc server-recv {i bytes} {
    $self instvar req_total_ req_bytes_
    incr req_bytes_($i) $bytes
    #puts "$self server-recv $i bytes: $bytes total: $req_bytes_($i)"
    if {$req_bytes_($i) == $req_total_($i)} {
	$self doneRequest $i
    }
}

Http instproc client-recv {i bytes} {
    $self instvar resp_total_ resp_bytes_ client_
    incr resp_bytes_($i) $bytes
    #puts "$self client-recv $i bytes: $bytes total: $resp_bytes_($i)"
    if {$resp_bytes_($i) == $resp_total_($i)} {
	[$client_($i) agent] close
    }
}

Http instproc client-done {i} {
    #puts "$self client-done $i"
    $self doneReply $i
}

Http instproc server-done {i} {
    $self instvar server_
    #puts "$self server-done $i"
    # reset the server so it listens again
    $server_($i) send 0
}

# Http::agents return the agents of type "source" or "sink"
Http instproc agents {{type source}} {
	$self instvar agents_
	return $agents_($type)
}


Http instproc client-transmit {i nbyte} {
    $self instvar req_total_ req_bytes_ client_ sessionLen_
    if {$nbyte < 40} {
	set nbyte 40
    }
    incr sessionLen_ $nbyte
    set req_total_($i) $nbyte
    set req_bytes_($i) 0
    $client_($i) send $nbyte
}

Http instproc server-transmit {i nbyte} {
    $self instvar resp_total_ resp_bytes_ server_ sessionLen_
    if {$nbyte < 40} {
	set nbyte 40
    }
    incr sessionLen_ $nbyte
    set resp_total_($i) $nbyte
    set resp_bytes_($i) 0
    $server_($i) send $nbyte
}


Http instproc start {} {
    $self instvar phttp_ maxConn_ rvClientTime_ rvServerTime_
    $self instvar rvReqLen_ rvRepLen_ rvNumImg_ rvImgLen_
    $self instvar ns_ numImg_ numGet_ numPut_ sessionLen_ \
	    client_ server_ tStart_
    $self instvar req_total_ req_bytes_ resp_total_ resp_bytes_
    $self instvar ct_ req_len_ tServer_ rep_len_ tnum_
    $self instvar rv_reqlen_tr_ rv_replen_tr_ rv_imglen_tr_ rv_imgnum_tr_ rv_ctime_tr_ rv_stime_tr_ 

    set tStart_ [$ns_ now]
    set numGet_ 0
    set numPut_ 0
    set sessionLen_ 0
#    set len [rvValue $rvReqLen_ round]
#    puts $rv_reqlen_tr_ $len
    
    # set all random vars for page at start so the order is deterministic
    #   and doesn't depend on the network setup
    set numImg_ [rvValue $rvNumImg_ round]
    puts $rv_imgnum_tr_ $numImg_
    set ct_ [rvValue $rvClientTime_]
    puts $rv_ctime_tr_ $ct_
    for {set i 0} {$i <= $numImg_} {incr i} {
	set req_len_($i) [rvValue $rvReqLen_ round]
	puts $rv_reqlen_tr_ $req_len_($i)
	set tServer_($i) [rvValue $rvServerTime_]
	puts $rv_stime_tr_ $tServer_($i)
	set rep_len_($i) [rvValue $rvRepLen_ round]
	puts $rv_replen_tr_ $rep_len_($i)
    }


	#puts "Http $self starts at $tStart_"
	#puts "$self reqH$numGet_ $client_(0) $len"

    set tnum_(0) 0
    $self client-transmit 0 $req_len_(0)

}

Http instproc doneRequest {id} {
	$self instvar phttp_ maxConn_ rvClientTime_ rvServerTime_
	$self instvar rvReqLen_ rvRepLen_ rvNumImg_ rvImgLen_
	$self instvar ns_ numImg_ numGet_ numPut_ client_ server_ tStart_ psz_ rv_replen_tr_ rv_stime_tr_
	$self instvar rep_len_ tServer_ tnum_

#	if {$numPut_ == 0} {
#		set len [rvValue $rvRepLen_ round]
#	} else {
#		set num [expr $phttp_ ? $numImg_ : 1]
#		for {set i [set len 0]} {$i < $num} {incr i} {
#			incr len [rvValue $rvImgLen_ round]
#		}
#	}

#	puts $rv_replen_tr_ $len

	set cmd "$self server-transmit $id $rep_len_($tnum_($id))"
#	set tServer [rvValue $rvServerTime_]
	set tServer $tServer_($tnum_($id))
	#puts "$self reply $server_($id) $len [$ns_ now] $tServer"
#	puts $rv_stime_tr_ $tServer
	if {$tServer > 0} {
#		$ns_ at [expr $tServer+[$ns_ now]] $cmd
		$ns_ after $tServer $cmd
	} else {
		eval $cmd
	}
}


Http instproc doneReply {id} {
	$self instvar phttp_ maxConn_ rvClientTime_ rvServerTime_
	$self instvar rvReqLen_ rvRepLen_ rvNumImg_ rvImgLen_
	$self instvar ns_ numImg_ numGet_ numPut_ client_ server_ tStart_ rv_imgnum_tr_ rv_reqlen_tr_
	$self instvar req_len_ tnum_

	set idlist $id
	if {$numPut_ == 0} {
		#set numImg_ [rvValue $rvNumImg_ round]
	        #puts $rv_imgnum_tr_ $numImg_
		for {set i 1} {$i < $maxConn_} {incr i} {
			lappend idlist $i
		}
	}
	if {[incr numPut_] > $numImg_} {
		$self donePage 
		return
	}

	# set numGet_ and numPut_ as if only 1 more request is needed
	foreach id $idlist {
		if {[incr numGet_] > $numImg_} break
#		if {$phttp_} {
#			set numGet_ $numImg_
#			set numPut_ $numImg_
#		}
                set tnum_($id) $numPut_
		#set len [rvValue $rvReqLen_ round]
		#puts $rv_reqlen_tr_ $len
		#puts "$self reqI$numGet_ $id $client_($id)  $req_len_($tnum_($id)) [$ns_ now]"
		$self client-transmit $id $req_len_($tnum_($id))
	}
}

Http instproc donePage {} {
	$self instvar phttp_ maxConn_ rvClientTime_ rvServerTime_ pagecount_ max_page_
	$self instvar ns_ numImg_ numGet_ numPut_ sessionLen_ \
			client_ server_ tStart_ rv_ctime_tr_ ct_
  $self instvar pagetime_tr_

   incr pagecount_

    set now [$ns_ now]
#    set ct [rvValue $rvClientTime_]
    set out [format "%.3f %.0f %d %.3f" [expr $now - $tStart_] [expr $numImg_+1] $sessionLen_ $ct_]
    set curtime [format "%.3f" $now]
    # print "curtime transaction_time num_connections sessionlen thinktime"
    puts "$pagecount_ $curtime Http $self donePage $out"
    puts $pagetime_tr_ [expr $now - $tStart_]

    if {$pagecount_ == $max_page_} {
	$ns_ at-now "finish"
    } else {
#        puts $rv_ctime_tr_ $ct
	$ns_ after  $ct_ "$self start"
    }
}









