
# Copyright (c) 2002 Massachusetts Institute of Technology,
# All rights reserved.
# 
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose, without fee, and without
# written agreement is hereby granted, provided that the above
# copyright notice and the following two paragraphs appear in all
# copies of this software.
# 
# IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE
# LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
# CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND
# ITS DOCUMENTATION, EVEN IF THE MASSACHUSETTS INSTITUTE OF
# TECHNOLOGY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY SPECIFICALLY DISCLAIMS
# ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
# SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE
# MASSACHUSETTS INSTITUTE OF TECHNOLOGY HAS NO OBLIGATION TO
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
# 
# Code contributed by Ronny Krashinsky, ronny@mit.edu, http://www.cag.lcs.mit.edu/~ronny


Class PSLink

set print_debug 0

PSLink set stay_alive_ 0
PSLink set stay_alive_time_ 0
PSLink set listen_interval_default_ 0
PSLink set listen_interval_max_ 8
PSLink set listen_interval_func_ li_null
PSLink set listen_interval_ratio_ 0.2
PSLink set beacon_period_ 0.1
PSLink set tr_base_ data/http
PSLink set trace_ 0

PSLink instproc stayalive {val} { 
    $self set stay_alive_time_ $val 
    $self set stay_alive_ [expr $val > 0]
}
PSLink instproc listen_interval_func {val} {$self set listen_interval_func_ $val}
PSLink instproc listen_interval_percent {val} {$self set listen_interval_ratio_ [expr $val / 100.0]}
PSLink instproc listen_interval_max {val} {$self set listen_interval_max_ $val}
PSLink instproc beacon_period_ms {val} { $self set beacon_period_ [expr $val / 1000.0] }
PSLink instproc tr_base {val} { $self set tr_base_ $val }
PSLink instproc trace {val} { $self set trace_ $val }

PSLink instproc init {ns mobile base_station args} {
    eval $self init-vars $args

    $self instvar mobile_ base_station_ ns_
    $self instvar mobi_q_ mobi_l_ base_q_ base_l_
    $self instvar beacon_listen_count_ beacon_skip_ 
    $self instvar listen_interval_ listen_interval_default_ listen_interval_max_ listen_interval_func_ beacon_period_ 
    $self instvar stay_alive_ stay_alive_time_ asleep_ mobi_idle_ base_idle_ delayed_sleep_count_
    $self instvar last_sleepwake_ total_sleep_time_ total_wake_time_ last_active_ reset_listen_interval_
    $self instvar tr_base_ pslink_tr_ state_tr_
    $self instvar trace_ wake_tr_ sleep_tr_ beacon_tr_ state_tr_

    set asleep_ 0
    set mobi_idle_ 1
    set base_idle_ 1
    set delayed_sleep_count_ 0
    
    set beacon_listen_count_ 0
    set beacon_skip_ 0
    set listen_interval_ $listen_interval_default_
    set reset_listen_interval_ 0
    set last_active_ 0

    set total_sleep_time_ 0
    set total_wake_time_ 0
    set last_sleepwake_ [$ns now]

    set pslink_tr_ [open "${tr_base_}_pslink.m" w]
#    set state_tr_ [open "${tr_base_}_state.dat" w]
    if {$trace_} {
	set wake_tr_ [open "${tr_base_}_wake.dat" w]
	set sleep_tr_ [open "${tr_base_}_sleep.dat" w]
	set beacon_tr_ [open "${tr_base_}_beacon.dat" w]
    }

    set ns_ $ns
    set mobile_ $mobile
    set base_station_ $base_station
 
    $ns_ duplex-link-op $mobile_ $base_station_ color green

    set mobi_q_ [$ns simplex-link-op $mobile $base_station queue]
    set mobi_l_ [$ns simplex-link-op $mobile $base_station link]
    set base_q_ [$ns simplex-link-op $base_station $mobile queue]
    set base_l_ [$ns simplex-link-op $base_station $mobile link]

    $mobi_q_ set-caught-sleeping-alert 1
    $mobi_l_ set-idle-alert 1
    $base_l_ set-idle-alert 1

    global print_debug
    if {$print_debug} { 
	puts "mobi_q: $mobi_q_"
	puts "base_q: $base_q_"
	puts "mobi_l: $mobi_l_"
	puts "base_l: $base_l_"
    }

    $mobi_q_ proc caught-sleeping-alert {} "$self caught-sleeping-alert $mobi_q_"
    $mobi_l_ proc idle-alert {idle} "$self idle-alert \$idle $mobi_l_"
    $base_l_ proc idle-alert {idle} "$self idle-alert \$idle $base_l_"
    

    $self sleep
    $ns_ after $beacon_period_ "$self beacon"
}


PSLink instproc finish {} {
    $self instvar total_sleep_time_ total_wake_time_ beacon_listen_count_ pslink_tr_ state_tr_
    $self instvar trace_ wake_tr_ sleep_tr_ beacon_tr_ state_tr_

    puts "pslink total sleep         : $total_sleep_time_"
    puts "pslink total wake          : $total_wake_time_"
    puts "pslink beacon-listen-count : $beacon_listen_count_"

    puts $pslink_tr_ "sleep = $total_sleep_time_;"
    puts $pslink_tr_ "wake  = $total_wake_time_;"
    puts $pslink_tr_ "beacons = $beacon_listen_count_;"

    close $pslink_tr_
#    close $state_tr_

    if {$trace_} {
	close $wake_tr_
	close $sleep_tr_
	close $beacon_tr_
    }


}

PSLink instproc caught-sleeping-alert {me} {
    $self instvar ns_ mobi_q_ base_q_ asleep_
    global print_debug
    if {$print_debug} { puts "$me has been caught sleeping! the time is [$ns_ now]" }
    
    if {$me == $mobi_q_} {
	$self wakeup
    }
}

PSLink instproc idle-alert {idle me} {
    $self instvar ns_ mobi_l_ base_l_ asleep_ mobi_idle_ base_idle_ reset_listen_interval_
    $self instvar stay_alive_ stay_alive_time_ active_ delayed_sleep_count_ last_active_ state_tr_

    if {$me == $mobi_l_} {
	set mobi_idle_ $idle
    }
    if {$me == $base_l_} {
	set base_idle_ $idle
    }

    if {$mobi_idle_ && $base_idle_} {
	set active_ 0
    } else {
	set active_ 1
    }

    global print_debug
    if {$print_debug} { puts "$me idle: $idle active: $active_ time: [$ns_ now]" }

#    if {! $active_} {
#	puts $state_tr_ "[$ns_ now] 1"
#    } elseif {! $mobi_idle_ && ! $base_idle_ } {
#	puts $state_tr_ "[$ns_ now] 4"
#    } elseif {! $mobi_idle_} {
#	puts $state_tr_ "[$ns_ now] 2"
#    } elseif {! $base_idle_} {
#	puts $state_tr_ "[$ns_ now] 3"
#    } else {
#	puts "state error"
#	exit
#    }


    if {$me == $mobi_l_} {
	if {$mobi_idle_} {
	    set last_active_ 0
	    set reset_listen_interval_ 1
	}
    }
    
    if {! $stay_alive_} {
	if {! $active_} {
	    $self sleep
	}
    } else {
	if {$me == $mobi_l_} {
	    if {$mobi_idle_} {
		incr delayed_sleep_count_
		if {$print_debug} { puts "delayed_sleep_count incr: $delayed_sleep_count_ time: [$ns_ now]" }
		$ns_ after $stay_alive_time_ "$self delayed-sleep"
	    }
	} elseif {$me == $base_l_} {
	    if {$delayed_sleep_count_ == 0 && ! $active_} {
		if {$print_debug} { puts "base done, go to sleep" }
		$self sleep
	    }
	}
    }

}

PSLink instproc delayed-sleep {} {
    $self instvar ns_ delayed_sleep_count_ active_
    #set delayed_sleep_count_ [expr $delayed_sleep_count_ - 1]
    incr delayed_sleep_count_ -1
    global print_debug
    if {$print_debug} { puts "delayed_sleep_count decr: $delayed_sleep_count_ time: [$ns_ now]" }
    if {$delayed_sleep_count_ == 0 && ! $active_} {
	$self sleep
    }
}

PSLink instproc beacon {} {
    $self instvar ns_ base_q_ asleep_ beacon_listen_count_ beacon_skip_ 
    $self instvar listen_interval_ last_active_ listen_interval_func_ beacon_period_
    $self instvar trace_ wake_tr_ sleep_tr_ beacon_tr_ state_tr_
    $self instvar total_wake_time_ total_sleep_time_ beacon_listen_count_

    $ns_ after $beacon_period_ "$self beacon"

    set bql [$base_q_ length]

    if {$trace_} {
	puts $wake_tr_ "$total_wake_time_"
	puts $sleep_tr_ "$total_sleep_time_"
	puts $beacon_tr_ "$beacon_listen_count_"
    }

    global print_debug
    if {$print_debug} { puts "beacon! time: [$ns_ now] bql: $bql beacon_skip: $beacon_skip_ listen_interval: $listen_interval_ last_active: $last_active_" }

    if {$asleep_} {
	if {$beacon_skip_ == $listen_interval_} {
	    if {$print_debug} { puts "listen!" }
	    #puts $state_tr_ "[$ns_ now] 5"
	    #puts $state_tr_ "[expr [$ns_ now] + 0.002] 0"
	    incr beacon_listen_count_
	    set beacon_skip_ 0
	    $self $listen_interval_func_
	    if {$bql > 0} {
		if {$print_debug} { puts "beacon, data waiting, time: [$ns_ now] bql: $bql" }
		$self wakeup
	    }
	} else {
	    incr beacon_skip_
	}
    }
    incr last_active_
}

PSLink instproc li_null {} {
}

PSLink instproc li_incr {} {
    $self instvar listen_interval_ listen_interval_max_
    incr listen_interval_
    if {$listen_interval_ > $listen_interval_max_} {
	set listen_interval_ $listen_interval_max_
    }
}

PSLink instproc li_double {} {
    $self instvar listen_interval_ listen_interval_max_
    if {$listen_interval_ == 0} {
	set listen_interval_ 1
    } else {
	set listen_interval_ [expr $listen_interval_ * 2]
    }
    if {$listen_interval_ > $listen_interval_max_} {
	set listen_interval_ $listen_interval_max_
    }
}

PSLink instproc li_percent {} {
    $self instvar listen_interval_ last_active_ listen_interval_max_ listen_interval_ratio_

    set listen_interval_ [expr floor($last_active_ * $listen_interval_ratio_) - 1]
    if {$listen_interval_ < 0} {
	set listen_interval_ 0
    }
    if {$listen_interval_ > $listen_interval_max_} {
	set listen_interval_ $listen_interval_max_
    }
}

PSLink instproc sleep {} {
    $self instvar ns_ mobi_q_ base_q_ asleep_ mobile_ base_station_ last_sleepwake_ total_wake_time_ state_tr_
    $self instvar beacon_skip_ listen_interval_ listen_interval_default_ reset_listen_interval_
    if {! $asleep_} {
	set timediff [expr [$ns_ now] - $last_sleepwake_]
	set total_wake_time_ [expr $total_wake_time_ + $timediff]
	set last_sleepwake_ [$ns_ now]
	global print_debug
	if {$print_debug} { puts "going to sleep! the time is [$ns_ now]" }
#	puts $state_tr_ "[$ns_ now] 0"
	set beacon_skip_ 0
	if {$reset_listen_interval_} {
	    set listen_interval_ $listen_interval_default_
	    set reset_listen_interval_ 0
	}
	$mobi_q_ sleep
	$base_q_ sleep
	set asleep_ 1
    }
}

PSLink instproc wakeup {} {
    $self instvar ns_ mobi_q_ base_q_ asleep_ mobile_ base_station_ last_sleepwake_ total_sleep_time_ state_tr_
    if {$asleep_} {
	set timediff [expr [$ns_ now] - $last_sleepwake_]
	set total_sleep_time_ [expr $total_sleep_time_ + $timediff]
	set last_sleepwake_ [$ns_ now]
	global print_debug
	if {$print_debug} { puts "time to wakeup! the time is [$ns_ now]" }
	$mobi_q_ wakeup
	$base_q_ wakeup
#	$ns_ after .000001 "$ns_ duplex-link-op $mobile_ $base_station_ color green" ;# delay to keep nam happy
	set asleep_ 0
    }
}







